/*----------------------------------------------------------------------------*/
/* physical model: generation of curtains                                     */
/*----------------------------------------------------------------------------*/

#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "beams.h"
#include "transform.h"

struct auroral_form rayed_arc_form = 
{
/* one-and-a-half period wave */
0., 1.5, 5.,
/* no folds */
0., 0., 0., 0.,
/* no jumps */
0., 0., 0.
};

struct auroral_form rayed_band_form = 
{
/* one-and-a-half period wave */
0., 1.5, 5.,
/* no folds */
0., 0., 0., 0.,
/* a couple of jumps */
0., 1.5, 5.
};

/*
 * TODO: this should include some bezier parameters for the folds, probably
 * at the moment i can't be arsed
 */
struct auroral_form corona_form = 
{
/* half period wave, low amplitude */
0., .5, 2.,
/* a couple of folds */
0., 3.5, 0.5, 0.,
/* no jumps */
0., 0., 0.
};

struct auroral_form drapery_form = 
{
/* one-and-a-half period wave */
0., 1.5, 5.,
/* loads of folds */
0., 7.5, 0.5, 0.,
/* no jumps */
0., 0., 0.
};

/*
 * input:
 * - n_points:    amount of points to generate
 * - start_vec:   position of first start point in curtain
 * - end_vec:     position of last start point in curtain
 * - mag_vec:     direction of the planet's magnetic field
 * - form:        auroral_form structure used to describe waves, folds and jumps
 *
 * output:
 * - points:      destination for generated points
 *
 * TODO: more input parameters?
 * TODO: folds using bezier
 */
void generate_curtain_startpoints(float *points, int n_points, float *start_vec, float *end_vec, float *mag_vec, struct auroral_form *form)
{
	int i;

	/* compute center direction vector.. */
	float dir_vec[N_DIM], p_vec[N_DIM];
	for (i=0; i<N_DIM; i++) 
		dir_vec[i] = (end_vec[i] - start_vec[i]) / ((float) n_points);

	/* compute perpendicular vector.. */
	p_vec[0]=mag_vec[1]*dir_vec[2] - mag_vec[2]*dir_vec[1];
	p_vec[1]=mag_vec[2]*dir_vec[0] - mag_vec[0]*dir_vec[2];
	p_vec[2]=mag_vec[0]*dir_vec[1] - mag_vec[1]*dir_vec[0];

	/* normalise the perpendicular vector.. */
	float r2 = 0;
	for (i=0; i<N_DIM; i++)
		r2 += p_vec[i]*p_vec[i];
	//printf("generate_curtain_startpoints(): r2 = %f, p_vec = (%f,%f,%f)\n", r2, p_vec[0], p_vec[1], p_vec[2]);
	for (i=0; i<N_DIM; i++)
		p_vec[i] /= sqrt(r2);
	//printf("generate_curtain_startpoints(): r2 = %f, p_vec = (%f,%f,%f)\n", r2, p_vec[0], p_vec[1], p_vec[2]);

	float d_phase = (form->wave_end_phase - form->wave_start_phase)*2.0*M_PI;
	float max_phase_step = d_phase / ((float) n_points);
	float iphase = form->wave_start_phase*2.0*M_PI;
	float phase;

	float x_center=start_vec[0], y_center=start_vec[1], z_center=start_vec[2];
	//printf("generate_curtain_startpoints(): center = (%f,%f,%f)\n", x_center, y_center, z_center);

	for (i=0; i<n_points; i++) {
		float x,y;

		float rnd_f = rand() / ((float) RAND_MAX);
		if (rnd_f >= .5) {
			phase = iphase + rnd_f*PHASE_DISPLACEMENT;
		} else {
			phase = iphase - rnd_f*PHASE_DISPLACEMENT;
		}

		float s = form->wave_amp * sin(phase);
		float xr = POS_DISPLACEMENT * form->wave_amp * rand() / ((float) RAND_MAX);
		float yr = POS_DISPLACEMENT * form->wave_amp * rand() / ((float) RAND_MAX);
		float zr = POS_DISPLACEMENT * form->wave_amp * rand() / ((float) RAND_MAX);
		*points++ = x_center + s*p_vec[0] + xr;
		*points++ = y_center + s*p_vec[1] + yr;
		*points++ = z_center + s*p_vec[2] + zr;

// 		*points++ = x_center + s*p_vec[0];
// 		*points++ = y_center + s*p_vec[1];
// 		*points++ = z_center + s*p_vec[2];


		#if 1
		/* increment center position with constant step.. */
		x_center += dir_vec[0];
		y_center += dir_vec[1];
		z_center += dir_vec[2];
		/* increment phase with random step.. */
		iphase += max_phase_step;
		#else
		/* increment center position with random step.. */
		if (rnd_f >= .5) {
			x_center += dir_vec[0] * rnd_f;
			y_center += dir_vec[1] * rnd_f;
			z_center += dir_vec[2] * rnd_f;
		} else {
			x_center -= dir_vec[0] * rnd_f;
			y_center -= dir_vec[1] * rnd_f;
			z_center -= dir_vec[2] * rnd_f;
		}
		/* increment phase with constant step.. */
		phase += max_phase_step;
		#endif

	}
}

void precipitate(float **dst_pt, float *start_point, float *mag_norm)
{

	float dt = 2./300.;
	/* vertical displacement */
	float t = 0;
	/* maximum vertical displacement (randomised a bit) */
	float t_max = 1.0 - .1*(rand() / ((float) RAND_MAX));
	/* 200 km max length? */
	float len = 200;
	/* position of collision, temporary position increase, definite (scattered) position increase */
	float p[N_DIM], vt[N_DIM], vp[N_DIM];
	/* iterator */
	int j;
	/* set up destination buffer pointer */
	float *dst;
	//printf("precipitate(): dst_pt=[%08X]\n", dst_pt);
	//printf("precipitate(): dst_pt=[%08X] *dst_pt=[%08X] **dst_pt=[%f]\n", dst_pt, *dst_pt, **dst_pt);
	dst = *dst_pt;
	//printf("precipitate()\n");

	float d_alpha = .015;
	/* = 3 deg */
	float alpha_d = 4.; //0.05235987755983066011; //.4

	memcpy(p, start_point, N_DIM*sizeof(float));
	#if 0
	for (j=0; j<N_DIM; j++) {
		printf("%f\t", p[j]);
	}
	printf("\n");
	#endif

	for (t=0; t<t_max; ) {
		float v = rand() / ((float) RAND_MAX);
		float dt_v = dt*v;
		//printf("dt*v=%f\n", dt_v);

		/* compute temporary (centered) displacement */
		for (j=0; j<N_DIM; j++) {
			vt[j] = dt_v*len*mag_norm[j];
			//printf("%f\t", vt[j]);
		}
		//printf("\n");

		/* compute actual displacement caused by deflection */
		float alpha = alpha_d*t*d_alpha;
		float beta = 2.*M_PI*rand() / ((float) RAND_MAX);
		/* TODO: this only works for one special magnetic field direction (0,0,z)! */
		vp[1] = cos(beta)*sin(alpha)*vt[2];
		vp[2] = cos(alpha)*vt[2];
		vp[0] = -sin(beta)*sin(alpha)*vt[2];

		/* compute position of next deflection */
		for (j=0; j<N_DIM; j++) {
			p[j] += vp[j];
			//printf("%f\t", p[j]);
			/* store coordinate.. */
			*dst++ = p[j];
		}
		//printf("\n");

		/* increase t with random step */
		t += dt_v;
	}

	/* copy back destination pointer */
	*dst_pt = dst;
	//printf("# precipitate(): dst=[%08X] dst_pt=[%08X] *dst_pt=[%08X] **dst_pt=[%f]\n", dst, dst_pt, *dst_pt, **dst_pt);
}
